@@ -303,7 +303,7 @@ partition_info ext_info;
 int num_parts = 0;
 
 int logical = 0;
-int logical_sectors[MAXIMUM_PARTS];
+long long logical_sectors[MAXIMUM_PARTS];
 
 __sighandler_t old_SIGINT, old_SIGTERM;
 
@@ -413,7 +413,7 @@ fdexit(int ret) {
 	fprintf(stderr, _("Reboot the system to ensure the partition "
 			"table is correctly updated.\n"));
 #endif
-	
+
 	fprintf( stderr, _("\nWARNING: If you have created or modified any\n"
 		         "DOS 6.x partitions, please see the cfdisk manual\n"
 		         "page for additional information.\n") );
@@ -1423,7 +1423,7 @@ new_part(int i) {
             { ESC, N_("Cancel"), N_("Don't create a partition") },
             { 0, NULL, NULL }
         };
-        
+
 	c = menuSimple( menuPartType, 0 );
 	if (toupper(c) == 'P')
 	    num = find_primary();
@@ -1685,6 +1685,7 @@ fill_p_info(void) {
 	 clear_warning();
     }
 
+#ifdef BLKFLSBUF
     /* Blocks are visible in more than one way:
        e.g. as block on /dev/hda and as block on /dev/hda3
        By a bug in the Linux buffer cache, we will see the old
@@ -1694,6 +1695,7 @@ fill_p_info(void) {
        so this only plays a role if we want to show volume labels. */
     ioctl(fd, BLKFLSBUF);	/* ignore errors */
 				/* e.g. Permission Denied */
+#endif
 
     if (blkdev_get_sectors(fd, &llsectors) == -1)
 	    fatal(_("Cannot get disk size"), 3);
@@ -1877,10 +1879,12 @@ write_part_table(void) {
 	      clear_warning();
 	      if (len == GS_ESCAPE)
 		   return;
-	      else if (strcasecmp(response, _("no")) == 0) {
+	      else if (strcasecmp(response, _("no")) == 0 ||
+		       strcasecmp(response, "no") == 0) {
 		   print_warning(_("Did not write partition table to disk"));
 		   return;
-	      } else if (strcasecmp(response, _("yes")) == 0)
+	      } else if (strcasecmp(response, _("yes")) == 0 ||
+			 strcasecmp(response, "yes") == 0)
 		   done = TRUE;
 	      else
 		   print_warning(_("Please enter `yes' or `no'"));
@@ -1903,10 +1907,12 @@ write_part_table(void) {
 	}
 
     if (is_bdev) {
+#ifdef BLKRRPART
 	 sync();
 	 sleep(2);
 	 if (!ioctl(fd,BLKRRPART))
 	      changed = TRUE;
+#endif
 	 sync();
 	 sleep(4);
 
@@ -1914,7 +1920,7 @@ write_part_table(void) {
 	 if (changed)
 	      print_warning(_("Wrote partition table to disk"));
 	 else
-	      print_warning(_("Wrote partition table, but re-read table failed.  Reboot to update table."));
+	      print_warning(_("Wrote partition table, but re-read table failed.  Run partprobe(8), kpartx(8) or reboot to update table."));
     } else
 	 print_warning(_("Wrote partition table to disk"));
 
@@ -2250,7 +2256,7 @@ print_tables(void) {
         { ESC, N_("Cancel"), N_("Don't print the table") },
         { 0, NULL, NULL }
     };
-    
+
     while (!done)
 	switch ( toupper(menuSimple( menuFormat, 2)) ) {
 	case 'R':
